from typing import Optional

from tests.payment_order.testdata import LE_INN
from vitya.error_description import AlertBody, AlertGenerator
from vitya.payment_order.enums import PaymentType
from vitya.payment_order.fields import (
    Amount,
    PayerAccountNumber,
    PayerINN,
    PayerKPP,
    PayerStatus,
    Receiver,
    ReceiverAccountNumber,
    ReceiverBIC,
    ReceiverKPP,
)
from vitya.payment_order.payments.checkers import BaseModelChecker


class Payment(BaseModelChecker):
    src_account: PayerAccountNumber  # [9]
    dst_account: ReceiverAccountNumber  # [17]
    amount: Amount  # [7]
    dst_name: Receiver  # [16]
    dst_bic: ReceiverBIC  # [14]
    dst_kpp: Optional[ReceiverKPP]  # [103]


ALERT_GENERATOR = AlertGenerator(
    {
        'amount': 'amount',
        'payer': 'dst_name',
        'receiver_bic': 'dst_bic',
        'receiver': 'dst_name',
        'receiver_account_number': 'dst_account',
    }
)


def test_alert_generator_not_none_field():
    try:
        Payment(
            src_account=PayerAccountNumber('40802810822200040036'),
            dst_account=ReceiverAccountNumber('40702810438000185552'),
            amount=None,
            dst_name=Receiver('Foo'),
            dst_bic=ReceiverBIC('044525225'),
        )
    except Exception as e:
        assert ALERT_GENERATOR.get_error_client_alerts(e) == [
            AlertBody(
                alert='Поле «Сумма» должно быть заполнено',
                failed_field='amount',
                failed_field_class_name='amount'
            )
        ]
    else:
        raise RuntimeError


def test_alert_generator_incorrect_len():
    try:
        Payment(
            src_account=PayerAccountNumber('40802810822200040036'),
            dst_account=ReceiverAccountNumber('40702810438000185552'),
            amount=Amount('1' * 20),
            dst_name=Receiver('Foo'),
            dst_bic=ReceiverBIC('044525225'),
        )
    except Exception as e:
        assert ALERT_GENERATOR.get_error_client_alerts(e) == [
            AlertBody(
                alert='Поле «Сумма» содержит неправильное количество символов',
                failed_field='amount',
                failed_field_class_name=None
            )
        ]
    else:
        raise RuntimeError


def test_alert_generator_exact_field_len():
    try:
        Payment(
            src_account=PayerAccountNumber('40802810822200040036'),
            dst_account='40702810438000185552-',
            amount=Amount('1'),
            dst_name=Receiver('Foo'),
            dst_bic=ReceiverBIC('044525225'),
        )
    except Exception as e:
        assert ALERT_GENERATOR.get_error_client_alerts(e) == [
            AlertBody(
                alert='Поле «Счет получателя» должно содержать ровно 20 символов',
                failed_field='receiver account',
                failed_field_class_name='dst_account'
            )
        ]
    else:
        raise RuntimeError


def test_alert_generator_incorrect_data():
    try:
        Payment(
            src_account=PayerAccountNumber('40802810822200040036'),
            dst_account=ReceiverAccountNumber('40702810438000185552'),
            amount=Amount('-1'),
            dst_name=Receiver('Foo'),
            dst_bic=ReceiverBIC('044525225'),
        )
    except Exception as e:
        assert ALERT_GENERATOR.get_error_client_alerts(e) == [
            AlertBody(
                alert='Поле «Сумма» содержит некорректные данные',
                failed_field='amount',
                failed_field_class_name=None
            )
        ]
    else:
        raise RuntimeError


class TestPayerKppChecker(BaseModelChecker):
    payer_kpp: Optional[PayerKPP]
    payment_type: PaymentType
    payer_inn: PayerINN
    payer_status: Optional[PayerStatus]


def test_alert_generator_need_required_field():
    try:
        TestPayerKppChecker(
            payer_kpp=None,
            payment_type=PaymentType.CUSTOMS,
            payer_inn=LE_INN,
            payer_status=None,
        )
    except Exception as e:
        assert ALERT_GENERATOR.get_error_client_alerts(e) == [
            AlertBody(
                alert='Поле «КПП плательщика» должно быть заполнено',
                failed_field='payer kpp',
                failed_field_class_name=None
            )
        ]
    else:
        raise RuntimeError



from pydantic.errors import PydanticTypeError, PydanticValueError

from vitya.errors_base import (
    ExactFieldLenError,
    IncorrectData,
    IncorrectLen,
    VityaDescribedError,
)


class OKTMOValidationError(VityaDescribedError, PydanticValueError):
    target = 'oktmo'
    target_ru = 'ОКТМО'
    description = 'base error'
    description_ru = 'базовая ошибка'


class OKTMOValidationTypeError(OKTMOValidationError, PydanticTypeError):
    description = 'must be str'
    description_ru = 'должен быть строкой'


class OKTMOValidationValueLenError(OKTMOValidationError, IncorrectLen):
    description = 'must be match 8 or 11 digits'
    description_ru = 'должен состоять из 8 или 11 цифр'


class OKTMOValidationValueError(OKTMOValidationError, IncorrectData):
    description = 'must be match as ([0-9]{11}|[0-9]{8})'
    description_ru = 'должен матчиться с ([0-9]{11}|[0-9]{8})'


class INNValidationError(VityaDescribedError, PydanticValueError):
    target = 'inn'
    target_ru = 'ИНН'
    description = 'base error'
    description_ru = 'базовая ошибка'


class INNValidationControlSumError(INNValidationError, IncorrectData):
    description = 'invalid control sum'
    description_ru = 'неверная контрольная сумма'


class INNValidationDigitsOnlyError(INNValidationError, IncorrectData):
    description = 'must contains only digits'
    description_ru = 'может содержать только цифры'


class INNValidationLenError(INNValidationError, IncorrectLen):
    description = 'len must be 5, 10 or 12'
    description_ru = 'длина должна быть 5, 10 или 12 символов'


class INNValidationTypeError(INNValidationError, PydanticTypeError):
    description = 'must be str'
    description_ru = 'должен быть строкой'


class INNValidationStartsWithZerosError(INNValidationError, IncorrectData):
    description = 'cannot starts with "00"'
    description_ru = 'не может начинаться с "00"'


class KPPValidationError(VityaDescribedError, PydanticValueError):
    target = 'kpp'
    target_ru = 'КПП'
    description = 'base error'
    description_ru = 'базовая ошибка'


class KPPValidationTypeError(KPPValidationError, PydanticTypeError):
    description = 'must be str'
    description_ru = 'должно быть строкой'


class KPPValidationValueLenError(KPPValidationError, ExactFieldLenError):
    description = 'kpp must be 9 chars'
    description_ru = 'должен содержать 9 символов'
    required_len = 9


class ReceiverKPPValidationValueDigitsOnlyError(KPPValidationError, IncorrectData):
    description = 'only digits allowed'
    description_ru = 'должен состоять только из цифр'
    target = 'receiver kpp'


class PayerKPPValidationValueDigitsOnlyError(KPPValidationError, IncorrectData):
    description = 'only digits allowed'
    description_ru = 'должен состоять только из цифр'
    target = 'payer kpp'


class KPPValidationValueError(KPPValidationError, IncorrectData):
    description = 'must matches as [0-9]{4}[0-9A-Z]{2}[0-9]{3}'
    description_ru = 'должен матчиться с [0-9]{4}[0-9A-Z]{2}[0-9]{3}'


class PayerKPPValidationValueCannotZerosStarts(KPPValidationError, IncorrectData):
    description = 'cannot starts with "00"'
    description_ru = 'не может начинаться с "00"'
    target = 'payer kpp'


class ReceiverKPPValidationValueCannotZerosStarts(KPPValidationError, IncorrectData):
    description = 'cannot starts with "00"'
    description_ru = 'не может начинаться с "00"'
    target = 'receiver kpp'


class BICValidationError(VityaDescribedError, PydanticValueError):
    target = 'bic'
    target_ru = 'БИК'
    description = 'base error'
    description_ru = 'базовая ошибка'


class BICValidationTypeError(BICValidationError, PydanticTypeError):
    description = 'must be str'
    description_ru = 'должен быть строкой'


class BICValidationLenError(BICValidationError, ExactFieldLenError):
    description = 'must be 9 chars'
    description_ru = 'должен содержать 9 символов'
    required_len = 9


class BICValidationValueDigitsOnlyError(BICValidationError):
    description = 'only digits allowed'
    description_ru = 'должен состоять только из чисел'



import re
from typing import List, Optional

from .errors import (
    BICValidationLenError,
    BICValidationTypeError,
    BICValidationValueDigitsOnlyError,
    INNValidationControlSumError,
    INNValidationDigitsOnlyError,
    INNValidationLenError,
    INNValidationStartsWithZerosError,
    INNValidationTypeError,
    KPPValidationTypeError,
    KPPValidationValueError,
    KPPValidationValueLenError,
    OKTMOValidationTypeError,
    OKTMOValidationValueError,
    OKTMOValidationValueLenError,
)


class ValidationError(ValueError):
    """
    Exception that raises if passed data is invalid
    """
    pass


def _count_inn_checksum(inn: str, coefficients: List[int]) -> int:
    assert len(inn) == len(coefficients)
    n = sum([int(digit) * coef for digit, coef in zip(inn, coefficients)])
    return n % 11 % 10


def validate_inn(inn: str, is_ip: Optional[bool] = None) -> Optional[str]:
    """
    Source:
    https://www.consultant.ru/document/cons_doc_LAW_134082/947eeb5630c9f58cbc6103f0910440cef8eaccac/
    https://ru.wikipedia.org/wiki/%D0%98%D0%B4%D0%B5%D0%BD%D1%82%D0%B8%D1%84%D0%B8%D0%BA%D0%B0%D1%86%D0%B8%D0%BE%D0%BD%D0%BD%D1%8B%D0%B9_%D0%BD%D0%BE%D0%BC%D0%B5%D1%80_%D0%BD%D0%B0%D0%BB%D0%BE%D0%B3%D0%BE%D0%BF%D0%BB%D0%B0%D1%82%D0%B5%D0%BB%D1%8C%D1%89%D0%B8%D0%BA%D0%B0
    """
    if not isinstance(inn, str):
        raise INNValidationTypeError

    if inn in {'', '0'}:
        return None

    if not re.fullmatch(r'[0-9]+', inn):
        raise INNValidationDigitsOnlyError

    if inn.startswith('00'):
        raise INNValidationStartsWithZerosError

    coefs10 = [2, 4, 10, 3, 5, 9, 4, 6, 8]
    coefs11 = [7] + coefs10
    coefs12 = [3] + coefs11

    if len(inn) == 10 and is_ip is not True:
        n10 = _count_inn_checksum(inn[:9], coefs10)
        if n10 != int(inn[9]):
            raise INNValidationControlSumError
        return inn
    elif len(inn) == 12 and is_ip is not False:
        n11 = _count_inn_checksum(inn[:10], coefs11)
        if n11 != int(inn[10]):
            raise INNValidationControlSumError

        n12 = _count_inn_checksum(inn[:11], coefs12)
        if n12 != int(inn[11]):
            raise INNValidationControlSumError
        return inn
    elif len(inn) == 5:
        return inn

    raise INNValidationLenError


def validate_inn_ip(inn: str) -> Optional[str]:
    return validate_inn(inn, is_ip=True)


def validate_inn_le(inn: str) -> Optional[str]:
    return validate_inn(inn, is_ip=False)


def validate_kpp(kpp: str) -> Optional[str]:
    """
    Source: https://kontur.ru/bk/spravka/491-chtotakoe_kpp
    """
    if not isinstance(kpp, str):
        raise KPPValidationTypeError
    elif kpp in {'0', ''}:
        return None

    if len(kpp) != 9:
        raise KPPValidationValueLenError

    if not re.fullmatch(r'[0-9]{4}[0-9A-Z]{2}[0-9]{3}', kpp):
        raise KPPValidationValueError
    return kpp


def validate_bic(bic: str) -> str:
    """
    Source:
    https://ru.wikipedia.org/wiki/%D0%91%D0%B0%D0%BD%D0%BA%D0%BE%D0%B2%D1%81%D0%BA%D0%B8%D0%B9_%D0%B8%D0%B4%D0%B5%D0%BD%D1%82%D0%B8%D1%84%D0%B8%D0%BA%D0%B0%D1%86%D0%B8%D0%BE%D0%BD%D0%BD%D1%8B%D0%B9_%D0%BA%D0%BE%D0%B4
    https://bik-info.ru/
    """
    if not isinstance(bic, str):
        raise BICValidationTypeError

    if len(bic) != 9:
        raise BICValidationLenError

    if not re.fullmatch(r'[0-9]+', bic):
        raise BICValidationValueDigitsOnlyError
    return bic


def validate_ogrn(ogrn: str, is_ip: Optional[bool] = None) -> None:
    """
    Source:
    https://ru.wikipedia.org/wiki/%D0%9E%D1%81%D0%BD%D0%BE%D0%B2%D0%BD%D0%BE%D0%B9_%D0%B3%D0%BE%D1%81%D1%83%D0%B4%D0%B0%D1%80%D1%81%D1%82%D0%B2%D0%B5%D0%BD%D0%BD%D1%8B%D0%B9_%D1%80%D0%B5%D0%B3%D0%B8%D1%81%D1%82%D1%80%D0%B0%D1%86%D0%B8%D0%BE%D0%BD%D0%BD%D1%8B%D0%B9_%D0%BD%D0%BE%D0%BC%D0%B5%D1%80
    """
    if not isinstance(ogrn, str):
        raise ValidationError('ogrn should be passed as string')

    if len(ogrn) != 13 and len(ogrn) != 15:
        raise ValidationError('wrong size of ogrn, it can be 13 chars only')

    if not re.fullmatch(r'[1-9][0-9]+', ogrn):
        raise ValidationError('wrong ogrn')

    if len(ogrn) == 13 and is_ip is not True:
        n13 = int(ogrn[:-1]) % 11 % 10
        if n13 != int(ogrn[12]):
            raise ValidationError(f'wrong checksum on pre-last digit: {ogrn[12]}; expected: {n13}')
        return

    if len(ogrn) == 15 and is_ip is not False:
        n15 = int(ogrn[:-1]) % 13 % 10
        if n15 != int(ogrn[14]):
            raise ValidationError(f'wrong checksum on pre-last digit: {ogrn[14]}; expected: {n15}')
        return

    raise ValidationError('ogrn for ip can be 15 chars only')


def validate_ogrnip(ogrnip: str) -> None:
    return validate_ogrn(ogrnip, is_ip=True)


def validate_snils(snils: str) -> None:
    """
    Source:
    https://www.consultant.ru/document/cons_doc_LAW_124607/68ac3b2d1745f9cc7d4332b63c2818ca5d5d20d0/
    """
    if not isinstance(snils, str):
        raise ValidationError('snils should be passed as string')

    if not re.fullmatch(r'[0-9]{11}', snils):
        raise ValidationError('wrong snils')

    if int(snils[:9]) < 1001998:  # less than 001-001-998
        raise ValidationError('snils must be more than "001-001-998" ')

    numbers = []
    parts = [snils[0:3], snils[3:6], snils[6:9]]
    for part in parts:
        numbers.extend([int(num) for num in part])

    results = [numbers[i - 1] * (10 - i) for i in range(1, 10)]
    checksum = sum(results)

    checksum = checksum % 101
    if checksum == 100:
        checksum_str = '00'
    elif checksum < 10:
        checksum_str = f'0{checksum}'
    else:
        checksum_str = str(checksum)

    if checksum_str != snils[-2:]:
        raise ValidationError(f'wrong checksum: {snils[-2:]}; expected: {checksum_str}')


def validate_oktmo(oktmo: str) -> Optional[str]:
    """
    Source:
    https://www.consultant.ru/cons/CGI/online.cgi?req=doc;base=LAW;n=149911#fUpVRbSdflobnNc4
    """
    if not isinstance(oktmo, str):
        raise OKTMOValidationTypeError
    elif oktmo in {'', '0'}:
        return None
    elif len(oktmo) not in {8, 11}:
        raise OKTMOValidationValueLenError
    if not re.fullmatch(r'([0-9]{11}|[0-9]{8})', oktmo):
        raise OKTMOValidationValueError
    return oktmo



from typing import Optional

import pytest
from pydantic import BaseModel, ValidationError as PydanticValidationError
from pydantic.errors import PydanticValueError

from vitya import (
    ValidationError as VityaValidationError,
    validate_bic,
    validate_inn,
    validate_inn_ip,
    validate_inn_le,
    validate_kpp,
    validate_ogrn,
    validate_ogrnip,
    validate_oktmo,
    validate_snils,
)
from vitya.errors import (
    OKTMOValidationTypeError,
    OKTMOValidationValueError,
    OKTMOValidationValueLenError,
)
from vitya.pydantic_fields import (
    BIC,
    INN,
    INNIP,
    INNLE,
    KPP,
    OGRN,
    OGRNIP,
    OKTMO,
    SNILS,
    FieldMixin,
)


class INNModel(BaseModel):
    inn: INN


class INNIPModel(BaseModel):
    inn: INNIP


class INNLEModel(BaseModel):
    inn: INNLE


class KPPModel(BaseModel):
    kpp: KPP


class BICModel(BaseModel):
    bic: BIC


class OGRNModel(BaseModel):
    ogrn: OGRN


class OGRNIPModel(BaseModel):
    ogrnip: OGRNIP


class SNILSModel(BaseModel):
    snils: SNILS


class OKTMOModel(BaseModel):
    oktmo: OKTMO


@pytest.mark.parametrize(
    'inn', [
        '3664069397', '302502032671', '7707083893', '7703206417', '771002344404'
    ]
)
def test_valid_inn(inn):
    """No exception raise"""
    validate_inn(inn)

    inn_model = INNModel(inn=inn)
    assert inn_model.inn == inn


@pytest.mark.parametrize(
    'inn', [
        '469933069430', '368332974449', '298410962506', '686899030369', '097289845404'
    ]
)
def test_valid_inn_ip(inn):
    validate_inn_ip(inn)

    inn_model_ip = INNIPModel(inn=inn)
    assert inn_model_ip.inn == inn

    with pytest.raises(PydanticValidationError):
        INNLEModel(inn=inn)


@pytest.mark.parametrize(
    'inn', [
        '9267145148', '5302008630', '6524062615', '0207895252', '0990471741'
    ]
)
def test_valid_inn_le(inn):
    validate_inn_le(inn)

    inn_model_le = INNLEModel(inn=inn)
    assert inn_model_le.inn == inn

    with pytest.raises(PydanticValidationError):
        INNIPModel(inn=inn)


@pytest.mark.parametrize(
    'inn', [
        None,  # can't be None
        3664069397,  # can't be nothing than str
        302502032671,
        '770708389',  # should be size of 10 or 12 chars
        '77100234440',
        '3664069398',  # wrong checksums
        '302502032672',
        '302502032681'
        '36640A9397'  # don't match regexp
    ]
)
def test_wrong_inn(inn):
    with pytest.raises(PydanticValueError):
        validate_inn(inn)

    with pytest.raises(PydanticValidationError):
        INNModel(inn=inn)

    with pytest.raises(PydanticValidationError):
        INNIPModel(inn=inn)

    with pytest.raises(PydanticValidationError):
        INNLEModel(inn=inn)


@pytest.mark.parametrize(
    'kpp', [
        '616401001', '770943002', '7709AB002', '320143522', '704601307',
    ]
)
def test_valid_kpp(kpp):
    """No exception raise"""
    validate_kpp(kpp)

    kpp_model = KPPModel(kpp=kpp)
    assert kpp_model.kpp == kpp


@pytest.mark.parametrize(
    'kpp', [
        None,  # can't be None
        616401001,  # can't be nothing than str
        770943002,
        '77070838',  # should be size of 9 chars
        '77100234440',
        '7709ABС02'  # don't match regexp
    ]
)
def test_wrong_kpp(kpp):
    with pytest.raises(PydanticValueError):
        validate_kpp(kpp)

    with pytest.raises(PydanticValidationError):
        KPPModel(kpp=kpp)


@pytest.mark.parametrize(
    'bic', [
        '044525901', '043002717', '046577964', '040349758', '041806647'
    ]
)
def test_valid_bic(bic):
    """No exception raise"""
    validate_bic(bic)

    bic_model = BICModel(bic=bic)
    assert bic_model.bic == bic


@pytest.mark.parametrize(
    'bic', [
        None,  # can't be None
        '',
        770943002,  # can't be nothing than str
        '04452590',  # should be size of 9 chars
        '0445259011',
        '034525901'  # don't match regexp
        '04452A901'
    ]
)
def test_wrong_bic(bic):
    with pytest.raises(PydanticValueError):
        validate_bic(bic)

    with pytest.raises(PydanticValidationError):
        BICModel(bic=bic)


@pytest.mark.parametrize(
    'ogrn', [
        '1027700132195', '1037700013020', '316784700262702', '304500116000157', '1076935620520'
    ]
)
def test_valid_ogrn(ogrn):
    """No exception raise"""
    assert validate_ogrn(ogrn) is None

    ogrn_model = OGRNModel(ogrn=ogrn)
    assert ogrn_model.ogrn == ogrn


@pytest.mark.parametrize(
    'ogrnip', [
        '304051927964808', '314057243354856', '307870729546242', '312550098407541', '308633624812989'
    ]
)
def test_valid_ogrnip(ogrnip):
    """No exception raise"""
    assert validate_ogrnip(ogrnip) is None

    ogrnip_model = OGRNIPModel(ogrnip=ogrnip)
    assert ogrnip_model.ogrnip == ogrnip


@pytest.mark.parametrize(
    'ogrnip', [
        '1076935620520', '5122703513136', '5081268440446', '5063675362394'
    ]
)
def test_wrong_ogrnip(ogrnip):
    with pytest.raises(PydanticValidationError):
        OGRNIPModel(ogrnip=ogrnip)


@pytest.mark.parametrize(
    'ogrn', [
        None,  # can't be None
        '',
        1027700132195,  # can't be nothing than str
        '102770013219',  # should be size of 13 or 15 chars
        '10377000130200',
        '10277001321955',
        '1027A00132195'  # don't match regexp
        '0027700132195',
        '1027700132196',  # wrong last digit
        '1037700013021'
        '304500116000158',
        '316784700262701'
    ]
)
def test_wrong_ogrn(ogrn):
    with pytest.raises(VityaValidationError):
        validate_ogrn(ogrn)

    with pytest.raises(PydanticValidationError):
        OGRNModel(ogrn=ogrn)

    with pytest.raises(PydanticValidationError):
        OGRNIPModel(ogrnip=ogrn)


@pytest.mark.parametrize(
    'snils', [
        '11223344595',
        '21647164763',
        '47789365577',
        '93149947849',
        '58966302961'
    ]
)
def test_valid_snils(snils):
    """No exception raise"""
    assert validate_snils(snils) is None

    snils_model = SNILSModel(snils=snils)
    assert snils_model.snils == snils


@pytest.mark.parametrize(
    'snils', [
        None,
        '',
        11223344595,
        '12-233-445 955',
        '12-233-445 9',
        '12-233-445-95',
        '12-233-44595',
        '12-233-445 96',
        '216-471-647 60',
        '00000000000'
    ]
)
def test_wrong_snils(snils):
    with pytest.raises(VityaValidationError):
        validate_snils(snils)

    with pytest.raises(PydanticValidationError):
        SNILSModel(snils=snils)


@pytest.mark.parametrize(
    'oktmo', [
        '69654000',
        '69701000001',
        '98603170051',
        '78623427116',
        '66614465117',
    ]
)
def test_valid_oktmo(oktmo):
    """No exception raise"""
    assert validate_oktmo(oktmo) == oktmo

    oktmo_model = OKTMOModel(oktmo=oktmo)
    assert oktmo_model.oktmo == oktmo


@pytest.mark.parametrize(
    'oktmo, error',
    [
        (None, OKTMOValidationTypeError),
        (69654000, OKTMOValidationTypeError),
        ('6965400', OKTMOValidationValueLenError),
        ('69b01000001', OKTMOValidationValueError),
    ]
)
def test_wrong_oktmo(oktmo, error):
    with pytest.raises(error):
        validate_oktmo(oktmo)

    with pytest.raises(PydanticValidationError):
        OKTMOModel(oktmo=oktmo)


class Field(FieldMixin, str):
    @classmethod
    def _validate(cls, value):
        return value if value not in {'', '0'} else None


class TestFieldMixin(BaseModel):
    field: Field


@pytest.mark.parametrize(
    'value',
    ('', '0', None)
)
def test_field_mixin(value: Optional[str]) -> None:
    with pytest.raises(PydanticValidationError):
        TestFieldMixin(field=value)


class TestFieldMixinOptional(BaseModel):
    field: Optional[Field]


@pytest.mark.parametrize(
    'value',
    ('', '0', None)
)
def test_field_mixin_optional(value: Optional[str]) -> None:
    assert TestFieldMixinOptional(field=value).field is None


class TestFieldMixinOptionalWithDefault(BaseModel):
    field: Optional[Field] = '5'


def test_field_mixin_optional_with_default() -> None:
    assert TestFieldMixinOptionalWithDefault().field == '5'


@pytest.mark.parametrize(
    'value',
    ('', '0', None),
)
def test_field_mixin_optional_with_default_with_value(value: Optional[str]) -> None:
    assert TestFieldMixinOptionalWithDefault(field=value).field is None



from abc import ABC, abstractmethod
from typing import Any, Callable, Generator, Optional

from pydantic.fields import ModelField

from .validators import (
    ValidationError,
    validate_bic,
    validate_inn,
    validate_inn_ip,
    validate_inn_le,
    validate_kpp,
    validate_ogrn,
    validate_ogrnip,
    validate_oktmo,
    validate_snils,
)

try:
    from pydantic.errors import MissingError, PydanticValueError
except ImportError:  # pragma: no cover
    pass

CallableGenerator = Generator[Callable[..., Any], None, None]


class PydanticValidationError(PydanticValueError):
    msg_template = 'invalid {name}: {reason}'


class EmptyError(Exception):
    pass


def _validate_wrapper(func: Callable[[str], None], name: str, value: str) -> str:
    try:
        func(value)
    except ValidationError as e:
        raise PydanticValidationError(name=name, reason=str(e))

    return value


class FieldMixin(ABC):
    def __new__(cls, value: Any) -> 'FieldMixin':
        if type(value) == cls:
            return value  # type: ignore
        value = cls._validate(value)
        if value is None:
            raise EmptyError
        return super().__new__(cls, value)  # type: ignore

    @classmethod
    @abstractmethod
    def _validate(cls, value: Any) -> Any:
        pass

    @classmethod
    def __get_validators__(cls) -> CallableGenerator:
        def validator(value: Any, field: ModelField) -> Any:
            try:
                return cls(value)
            except EmptyError:
                if field.allow_none:
                    return None
                raise MissingError
        yield validator


class INN(FieldMixin, str):
    @classmethod
    def _validate(cls, value: str) -> Optional[str]:
        return validate_inn(value)


class INNIP(FieldMixin, str):
    @classmethod
    def _validate(cls, value: str) -> Optional[str]:
        return validate_inn_ip(value)


class INNLE(FieldMixin, str):
    @classmethod
    def _validate(cls, value: str) -> Optional[str]:
        return validate_inn_le(value)


class KPP(FieldMixin, str):
    @classmethod
    def _validate(cls, value: str) -> Optional[str]:
        return validate_kpp(value)


class BIC(FieldMixin, str):
    @classmethod
    def _validate(cls, value: str) -> str:
        return validate_bic(value)


class OGRN(FieldMixin, str):
    @classmethod
    def _validate(cls, value: str) -> str:
        return _validate_wrapper(validate_ogrn, 'ogrn', value)


class OGRNIP(FieldMixin, str):
    @classmethod
    def _validate(cls, value: str) -> str:
        return _validate_wrapper(validate_ogrnip, 'ogrn_ip', value)


class SNILS(FieldMixin, str):
    @classmethod
    def _validate(cls, value: str) -> str:
        return _validate_wrapper(validate_snils, 'snils', value)


class OKTMO(FieldMixin, str):
    @classmethod
    def _validate(cls, value: str) -> Optional[str]:
        return validate_oktmo(value)


# Pydantic strip NewType from ModelField.type_

class BoolWrapper:
    __slots__ = ('_value', )

    def __init__(self, value: bool):
        self._value = value

    def __bool__(self) -> bool:
        return self._value

    def __hash__(self) -> int:
        return hash(self._value)

    def __eq__(self, other: Any) -> Any:
        if isinstance(other, type(self)):
            return self._value == other._value
        if isinstance(other, bool):
            return self._value == other
        return NotImplemented

    @classmethod
    def __get_validators__(cls) -> CallableGenerator:
        def validator(value: Any) -> Any:
            if isinstance(value, bool):
                return cls(value)
            if isinstance(value, cls):
                return value
            raise TypeError

        yield validator



from decimal import Decimal
from typing import Any, Callable, Generator, Optional

from vitya.payment_order.validators import (
    validate_account_number,
    validate_amount,
    validate_cbc,
    validate_customer,
    validate_document_date,
    validate_document_number,
    validate_number,
    validate_operation_kind,
    validate_payer,
    validate_payer_kpp,
    validate_payer_status,
    validate_payment_order,
    validate_purpose,
    validate_purpose_code,
    validate_reason,
    validate_receiver,
    validate_receiver_account_number,
    validate_receiver_kpp,
    validate_tax_period,
    validate_type_of_income,
    validate_uin,
)
from vitya.pydantic_fields import BIC, INN, KPP, BoolWrapper, FieldMixin

CallableGenerator = Generator[Callable[..., Any], None, None]


class Number(FieldMixin, str):
    """Номер платёжного поручения (3)"""

    @classmethod
    def _validate(cls, value: str) -> str:
        return validate_number(value)


class Amount(FieldMixin, Decimal):
    """Сумма (7)"""

    @classmethod
    def _validate(cls, value: str) -> Decimal:
        return validate_amount(value)


class Customer(FieldMixin, str):
    """Общий класс для описания сущности владельца денег"""

    @classmethod
    def _validate(cls, value: str) -> str:
        return validate_customer(value)


class Payer(Customer):
    """Плательщик (8)"""

    @classmethod
    def _validate(cls, value: str) -> str:
        return validate_payer(value)


class Receiver(Customer):
    """Наименование получателя (16)"""

    @classmethod
    def _validate(cls, value: str) -> str:
        return validate_receiver(value)


class PaymentOrder(FieldMixin, int):
    """Очерёдность платежа (21)"""

    @classmethod
    def _validate(cls, value: str) -> int:
        return validate_payment_order(value)


class AccountNumber(FieldMixin, str):
    """Счёт"""

    @classmethod
    def _validate(cls, value: str) -> str:
        return validate_account_number(value)


class ReceiverAccountNumber(AccountNumber):
    """Счет получателя (17)"""

    @classmethod
    def _validate(cls, value: str) -> str:
        return validate_receiver_account_number(value)


class PayerAccountNumber(AccountNumber):
    """Номер счёта плательщика (9)"""


class OperationKind(FieldMixin, str):
    """Вид операции (18)"""

    @classmethod
    def _validate(cls, value: str) -> str:
        return validate_operation_kind(value)


class UIN(FieldMixin, str):
    """Код (УИН) (22)"""

    @classmethod
    def _validate(cls, value: str) -> Optional[str]:
        return validate_uin(value)


class PurposeCode(FieldMixin, int):
    """Назначение платежа кодовое (20)"""

    @classmethod
    def _validate(cls, value: int) -> int:
        return validate_purpose_code(value)


class Purpose(FieldMixin, str):
    """Назначение платежа (24)"""

    @classmethod
    def _validate(cls, value: str) -> Optional[str]:
        return validate_purpose(value)


class PayerINN(INN):
    """ИНН плательщика (60)"""


class ReceiverINN(INN):
    """ИНН получателя (61)"""


class ReceiverBIC(BIC):
    """БИК банка получателя (14)"""


class ReceiverKPP(KPP):
    """КПП плательщика (103)"""

    @classmethod
    def _validate(cls, value: str) -> Optional[str]:
        return validate_receiver_kpp(value)


class PayerKPP(KPP):
    """КПП получателя (102)"""

    @classmethod
    def _validate(cls, value: str) -> Optional[str]:
        return validate_payer_kpp(value)


class PayerStatus(FieldMixin, str):
    """Статус плательщика (101)"""

    @classmethod
    def _validate(cls, value: str) -> str:
        return validate_payer_status(value)


class CBC(FieldMixin, str):
    """КБК (104)"""

    @classmethod
    def _validate(cls, value: str) -> Optional[str]:
        return validate_cbc(value)


class Reason(FieldMixin, str):
    """Основание платежа (106)"""

    @classmethod
    def _validate(cls, value: str) -> Optional[str]:
        return validate_reason(value)


class TaxPeriod(FieldMixin, str):
    """Периодичность платежа / Код таможенного органа (107)"""

    @classmethod
    def _validate(cls, value: str) -> Optional[str]:
        return validate_tax_period(value)


class DocumentNumber(FieldMixin, str):
    """Номер документа (108)"""

    @classmethod
    def _validate(cls, value: str) -> Optional[str]:
        return validate_document_number(value)


class DocumentDate(FieldMixin, str):
    """Дата документа (109)"""

    @classmethod
    def _validate(cls, value: str) -> Optional[str]:
        return validate_document_date(value)


class ForThirdPerson(BoolWrapper):
    pass


class TypeOfIncome(FieldMixin, str):
    """Код вида дохода"""

    @classmethod
    def _validate(cls, value: str) -> Optional[str]:
        return validate_type_of_income(value)



from enum import Enum
from typing import List


class AccountKind(str, Enum):
    IP = 'ip'                           # счет ИП
    FL = 'fl'                           # счет ФЛ
    LE = 'le'                           # счет ЮР
    CHAMELEON = 'chameleon'             # счет Хамелеона


class PaymentType(str, Enum):
    FNS = 'fns'                         # бюджетный платеж в ФНС
    CUSTOMS = 'customs'                 # бюджетный платеж в таможню
    BUDGET_OTHER = 'budget_other'       # бюджетный платеж иное
    IP = 'ip'                           # платеж на ИП
    FL = 'fl'                           # платеж на ФЛ
    LE = 'le'                           # платеж на ЮР
    CHAMELEON = 'chameleon'             # платеж на Хамелеона

    @classmethod
    def budget_types(cls) -> List['PaymentType']:
        return [PaymentType.FNS, PaymentType.CUSTOMS, PaymentType.BUDGET_OTHER]

    @property
    def is_budget(self) -> bool:
        return self in self.budget_types()

    @property
    def name_ru(self) -> str:
        return _PAYMENT_TYPE_TO_RU[self]


_PAYMENT_TYPE_TO_RU = {
    PaymentType.FNS: 'ФНС',
    PaymentType.CUSTOMS: 'Таможня',
    PaymentType.BUDGET_OTHER: 'Иные',
    PaymentType.IP: 'ИП',
    PaymentType.FL: 'ФЛ',
    PaymentType.LE: 'ЮЛ',
    PaymentType.CHAMELEON: 'Хамелеон',
}



from datetime import date
from typing import Any

from pydantic.errors import PydanticTypeError, PydanticValueError

from vitya.errors import (
    INNValidationError,
    INNValidationLenError,
    KPPValidationError,
    OKTMOValidationError,
)
from vitya.errors_base import (
    ExactFieldLenError,
    IncorrectData,
    IncorrectLen,
    NeedRequiredField,
    VityaDescribedError,
)
from vitya.payment_order.enums import PaymentType
from vitya.payment_order.payments.constants import (
    CHANGE_YEAR,
    CUSTOMS_REASONS,
    DOCUMENT_NUMBERS,
    FNS_KPP,
    FTS_KPP,
    FTS_OKTMO,
    PAYER_STATUSES,
    PAYER_STATUSES_AFTER_2024,
)


class PaymentTypeValueError(VityaDescribedError, PydanticValueError):
    target = 'payment type'
    target_ru = 'Тип платежа'
    description = 'неизвестный тип платежа '
    description_ru = 'неизвестный тип платежа '

    def __init__(self, payment_type: PaymentType, *args: Any, **kwargs: Any) -> None:  # pragma: no cover
        super().__init__(*args, **kwargs)
        self.description += payment_type.name
        self.description_ru += payment_type.name_ru


class AmountValidationError(VityaDescribedError, PydanticValueError):
    target = 'amount'
    target_ru = 'Сумма'
    description = 'base error'
    description_ru = 'базовая ошибка'


class AmountValidationLengthError(AmountValidationError, IncorrectLen):
    description = 'str representation cannot contains more 18 chars'
    description_ru = 'строковое представление не может содержать более 18 символов'


class AmountValidationLessOrEqualZeroError(AmountValidationError, IncorrectData):
    description = 'cannot be less than or equal to 0.0'
    description_ru = 'не может быть меньше или равно 0.0'


class AmountNotANumber(AmountValidationError, IncorrectData):
    description = 'require to be a number'
    description_ru = 'должно быть числом'


class CustomerValidationError(VityaDescribedError, PydanticValueError):
    target = 'customer'
    target_ru = 'Плательщик или Получатель'
    description = 'base error'
    description_ru = 'базовая ошибка'


class CustomerValidationSizeError(CustomerValidationError):
    description = 'name must sized from 1 to 160 chars'
    description_ru = 'длина имени должно быть между 1 и 160 символами'


class PayerValidationError(CustomerValidationError):
    target = 'payer'
    target_ru = 'Плательщик'


class PayerValidationSizeError(PayerValidationError, CustomerValidationSizeError, IncorrectLen):
    pass


class ReceiverValidationError(CustomerValidationError):
    target = 'receiver'
    target_ru = 'Получатель'


class ReceiverValidationSizeError(ReceiverValidationError, CustomerValidationSizeError, IncorrectLen):
    pass


class ReceiverValidationNameError(ReceiverValidationError, IncorrectData):
    description = 'contains account number'
    description_ru = 'содержит номер счета'


class NumberValidationLenError(PydanticValueError, IncorrectLen):
    target = 'number'
    target_ru = 'Номер'
    description = 'cannot be longer than 6 chars'
    description_ru = 'не может быть длиннее 6 символов'


class PaymentOrderValidationError(VityaDescribedError, PydanticValueError):
    target = 'payment order'
    target_ru = 'Очередность платежа'
    description = 'value must be in {1, 2, 3, 4, 5}'
    description_ru = 'значение должно быть одним из {1, 2, 3, 4, 5}'


class PaymentOrderLenError(PaymentOrderValidationError, ExactFieldLenError):
    required_len = 1


class OperationKindValidationError(VityaDescribedError, PydanticValueError):
    target = 'operation kind'
    target_ru = 'Вид операции'
    description = 'base error'
    description_ru = 'базовая ошибка'


class OperationKindValidationTypeError(OperationKindValidationError, PydanticTypeError):
    description = 'must be str'
    description_ru = 'должен быть строкой'


class OperationKindValidationBudgetValueError(OperationKindValidationError):
    description = 'ior budget must be one of {"01", "02", "06"}'
    description_ru = 'для бюджетных операций значение должно быть одним из {"01", "02", "06"}'


class OperationKindValidationValueError(OperationKindValidationBudgetValueError, ExactFieldLenError):
    description = 'value must be str with len 2'
    description_ru = 'значение должно быть длиной 2 символов'
    required_len = 2


class PurposeCodeValidationError(VityaDescribedError, PydanticValueError):
    target = 'purpose code'
    target_ru = 'Код назначения'
    description = 'base error'
    description_ru = 'базовая ошибка'


class PurposeCodeValidationTypeError(PurposeCodeValidationError, PydanticTypeError):
    description = 'must be int'
    description_ru = 'должен быть числом'


class PurposeCodeValidationNullError(PurposeCodeValidationError):
    description = 'for non fl payment value must be null'
    description_ru = 'должен отсутствовать для не ФЛ платежей и не хамелеона'


class PurposeCodeValidationFlError(PurposeCodeValidationError):
    description = 'for fl payment value must be in {1, 2, 3, 4, 5} or must be empty'
    description_ru = 'для платежей ФЛ должен быть одним из {1, 2, 3, 4, 5} или быть пустым'


class PurposeCodeValidationChameleonError(PurposeCodeValidationError):
    description = 'for chameleon payment value must be in {1, 2, 3, 4, 5} or must be empty'
    description_ru = 'для платежей хамелеону должен быть одним из {1, 2, 3, 4, 5} или быть пустым'


class PurposeValidationForThirdPersonError(PurposeCodeValidationError):
    description = 'for third person purpose must match pattern inn//name//purpose'
    description_ru = 'для платежей за третьих лиц назначение должно соответствовать шаблону ИНН//ФИО//Назначение'


class UINValidationError(VityaDescribedError, PydanticValueError):
    target = 'uin'
    target_ru = 'УИН'
    description = 'base error'
    description_ru = 'базовая ошибка'


class UINValidationTypeError(UINValidationError, PydanticTypeError):
    description = 'must be str'
    description_ru = 'должен быть строкой'


class UINValidationLenError(UINValidationError, IncorrectLen):
    description = 'len must be 4, 20 or 25 len'
    description_ru = 'длина должна быть 4, 20 или 25 символов'


class UINValidationDigitsOnlyError(UINValidationError, IncorrectData):
    description = 'must contains only digits'
    description_ru = 'должен содержать только числа'


class UINValidationControlSumError(UINValidationError, IncorrectData):
    description = 'control sum error'
    description_ru = 'некорректная контрольная сумма'


class UINValidationValueZeroError(UINValidationError, NeedRequiredField):
    description = 'value cannot be zero'
    description_ru = 'значение не может быть нулем'


class UINValidationValueBudget33PayerStatusIncorrectLength(UINValidationError, NeedRequiredField):
    description = 'value cannot consist of zeros and be of any length aprt from 20 or 25 if payer status is 33'
    description_ru = 'значение не может состоять из нулей и быть любой длины кроме 20 или 25 если статус плательщика 33'


class UINValidationBOLenError(UINValidationLenError):
    description = 'len uin for bo payment must be 4, 20 or 25 len'
    description_ru = 'длина для платежей с типом иные должна быть 4, 30 или 25'


class UINValidationFNSValueError(UINValidationError):
    description = 'FNS base error'
    description_ru = 'базовая ФНС ошибка'


class UINValidationFNSValueZeroError(UINValidationFNSValueError, NeedRequiredField):
    description = (
        'for FNS with payer status = "13" and empty inn uin must be non zero'
    )
    description_ru = (
        'для платежей в ФНС со статусом плательщика "13" и пустым ИНН, УИН должен быть не нулевым'
    )


class UINValidationFNSNotValueZeroError(UINValidationFNSValueError, IncorrectData):
    description = 'for FNS with payer status = "02" uin must be zero'
    description_ru = 'для платежей в ФНС со статусом плательщика "02" и пустым ИНН, УИН должен быть нулевым'


class UINValidationBONotEmpty(UINValidationFNSValueError, NeedRequiredField):
    description = 'for budget other and receiver number starting with 03212 uin must be filled'
    description_ru = (
        'для платежей в бюджет иные с номером счёта получателя, который начинается с "03212", УИН должен быть заполен'
    )


class UINValidationFNSOrFTSLenError(UINValidationLenError):
    description = 'invalid uin: len uin for FNS or FTS payment must be 20 or 25 len'
    description_ru = 'длина УИН для платежей в ФНС или ФТС должна быть 20 или 15 символов'


class UINValidationOnlyZeroError(UINValidationError, IncorrectData):
    description = 'cannot contains only 0 chars'
    description_ru = 'не может состоять только из нулей'


class PurposeValidationError(VityaDescribedError, PydanticValueError):
    target = 'purpose'
    target_ru = 'Назначение'
    description = 'base error'
    description_ru = 'базовая ошибка'


class PurposeValidationTypeError(PurposeValidationError, PydanticTypeError):
    description = 'must be str'
    description_ru = 'должно быть строкой'


class PurposeValidationMaxLenError(PurposeValidationError, IncorrectLen):
    description = 'len can be from 1 to 210 chars'
    description_ru = 'длина должна быть от 1 до 210 символов'


class PurposeValidationValueEmptyErrorForNonFNS(PurposeValidationError, NeedRequiredField):
    description = 'value cannot be empty for non FNS payments'
    description_ru = 'значение не может быть пустым для платежей не в ФНС'


class PayerINNValidationError(INNValidationError):
    target = 'payer inn'
    target_ru = 'ИНН плательщика'


class PayerINNValidationCustomsLen10Error(INNValidationLenError, PayerINNValidationError, ExactFieldLenError):
    description = 'for customs payment and payer status "06", inn must be 10'
    description_ru = 'для платежей в таможню и со статусом плательщика "06" ИНН должно быть быть длиной 10 символов'
    required_len = 10


class PayerINNValidationCustomsLen12Error(INNValidationLenError, PayerINNValidationError, ExactFieldLenError):
    description = 'for customs payment and payer status 16 or 17, inn must be 12'
    description_ru = (
        'для платежей таможню со статусом плательщика "16" или "17",'
        ' значение ИНН должно быть длиной 12 символов'
    )
    required_len = 12


class PayerINNValidationEmptyNotAllowedError(PayerINNValidationError, NeedRequiredField):
    description = 'inn cannot be empty'
    description_ru = 'не может быть пустым'


class PayerINNValidationStartWithZerosError(PayerINNValidationError, IncorrectData):
    description = 'cannot start with "00"'
    description_ru = 'не может начинаться с "00"'


class ReceiverINNValidationError(INNValidationError):
    target = 'receiver inn'
    target_ru = 'ИНН получателя'


class ReceiverINNValidationNonEmptyError(ReceiverINNValidationError, NeedRequiredField):
    description = 'cannot be empty'
    description_ru = 'не может быть пустым'


class ReceiverINNValidationFLenError(ReceiverINNValidationError, ExactFieldLenError):
    required_len = 12
    description = 'for fl receiver inn must be 12'
    description_ru = 'для платежей ИП ИНН получателя должно быть длиной 12 символов'


class ReceiverINNValidationFLLenError(ReceiverINNValidationError, ExactFieldLenError):
    required_len = 12
    description = 'for fl receiver inn must be empty or 12 chars'
    description_ru = 'для платежей ИП ИНН получателя должно быть пустым или содержать 12 символов'


class ReceiverINNValidationChameleonLenError(INNValidationLenError, ReceiverINNValidationError):
    description = 'for chameleon receiver inn must be empty or 12 or 10 chars'
    description_ru = 'для платежей Хамелеону ИНН получателя должно быть пустым или содержать 12 или 10 символов'


class ReceiverINNValidationIPLenError(ReceiverINNValidationError, ExactFieldLenError):
    required_len = 12
    description = 'for ip receiver inn must be 12'
    description_ru = 'для платежей ИП ИНН получателя должно быть 12 символов'


class ReceiverINNValidationLELenError(ReceiverINNValidationError, ExactFieldLenError):
    required_len = 10
    description = 'for fns, customs, bo and le inn must be 10'
    description_ru = 'для платежей в бюджет и платежей ЮЛ, ИНН должно быть 10 символов'


class ReceiverAccountValidationError(VityaDescribedError, PydanticValueError):
    target = 'receiver account'
    target_ru = 'Счет получателя'
    description = 'base error'
    description_ru = 'базовая ошибка'


class ReceiverAccountValidationNonEmptyError(ReceiverAccountValidationError):
    description = 'cannot be empty'
    description_ru = 'не может быть пустым'


class ReceiverAccountValidationLenError(ReceiverAccountValidationError):
    description = 'must be 20 digits'
    description_ru = 'должен состоять из 20 цифр'


class ReceiverAccountValidationFNSValueError(ReceiverAccountValidationError, IncorrectData):
    description = 'for FNS payment account must be "03100643000000018500"'
    description_ru = 'для платежей в ФНС счет должен быть "03100643000000018500"'


class AccountNumberValidationError(VityaDescribedError, PydanticValueError):
    target = 'account number'
    target_ru = 'Номер счета'
    description = 'base error'
    description_ru = 'базовая ошибка'


class AccountNumberValidationTypeError(AccountNumberValidationError, PydanticTypeError):
    description = 'must be str'
    description_ru = 'должен быть строкой'


class AccountNumberValidationSizeError(AccountNumberValidationError, ExactFieldLenError):
    description = 'must consist of 20 chars'
    description_ru = 'должен состоять из 20 символов'
    required_len = 20


class AccountNumberValidationDigitsOnlyError(AccountNumberValidationError):
    description = 'must be only digits'
    description_ru = 'должен состоять только из цифр'


class AccountValidationBICValueError(AccountNumberValidationError):
    description = 'account is not valid for bic'
    description_ru = 'не ключуется с БИКом'


class ReceiverAccountNumberValidationTypeError(
    ReceiverAccountValidationError,
    AccountNumberValidationTypeError,
):
    pass


class ReceiverAccountNumberValidationSizeError(
    ReceiverAccountValidationError,
    AccountNumberValidationSizeError,
):
    pass


class ReceiverAccountNumberValidationDigitsOnlyError(
    ReceiverAccountValidationError,
    AccountNumberValidationDigitsOnlyError,
    IncorrectData
):
    pass


class ReceiverAccountValidationBICValueError(
    ReceiverAccountValidationError,
    AccountValidationBICValueError,
    IncorrectData,
):
    pass


class ReceiverAccountValidationCustomsValueError(
    ReceiverAccountValidationError,
    IncorrectData
):
    description = 'for customs payment account must be "03100643000000019502"'
    description_ru = 'для платежей в таможню счет должен быть "03100643000000019502"'


class PayerStatusValidationError(VityaDescribedError, PydanticValueError):
    target = 'payer status'
    target_ru = 'Статус плательщика'
    description = 'base error'
    description_ru = 'базовая ошибка'


class PayerStatusValidationTypeError(PayerStatusValidationError, PydanticTypeError):
    description = 'must be str'
    description_ru = 'должен быть строкой'


class PayerStatusValidationValueError(PayerStatusValidationError):
    description = f'value can be only {PAYER_STATUSES if date.today().year < CHANGE_YEAR else PAYER_STATUSES_AFTER_2024}'
    description_ru = f'должен быть одним из {PAYER_STATUSES if date.today().year < CHANGE_YEAR else PAYER_STATUSES_AFTER_2024}'


class PayerStatusValidationNullNotAllowedError(PayerStatusValidationError):
    description = 'for budget payments empty value is not allowed'
    description_ru = 'для платежей в бюджет значение не должно быть пустым'


class PayerStatusValidationCustoms05NotAllowedError(PayerStatusValidationError, IncorrectData):
    description = 'for customs payment and for_third_face = true value "06" not allowed'
    description_ru = 'для платежей в таможню и для платежей за третьих лиц значение статуса не может быть "06"'


class PayerStatusValidationCustomsIncorrectDataError(PayerStatusValidationError, IncorrectData):
    description = 'for customs payment only values "06", "16", "17", "28", "30" are allowed'
    description_ru = 'для платежей в таможню значение статуса может быть только "06", "16", "17", "28", "30"'


class PayerStatusValidationFNSIncorrectDataError(PayerStatusValidationError, IncorrectData):
    description = 'for fns payment only values "01", "13" are allowed'
    description_ru = 'для платежей в налоговую значение статуса может быть только "01", "13"'


class PayerStatusValidationOtherIncorrectDataError(PayerStatusValidationError, IncorrectData):
    description = 'for other payment values "01", "13", "06", "16", "17", "28", "30" not allowed'
    description_ru = (
        'для платежей в иные организации значение статуса не может быть "01", "13", "06", "16", "17", "28", "30"'
    )


class KPPValidationOnlyEmptyError(KPPValidationError):
    description = 'only empty is allowed'
    description_ru = 'должно быть пустым'


class KPPValidationEmptyNotAllowed(KPPValidationError):
    description = 'empty value is not allowed'
    description_ru = 'пустое значение не разрешено'


class PayerKPPValidationError(KPPValidationError):
    target = 'payer kpp'
    target_ru = 'КПП плательщика'
    description = 'base error'
    description_ru = 'базовая ошибка'


class PayerKPPValidationOnlyEmptyError(PayerKPPValidationError, KPPValidationOnlyEmptyError):
    description = 'for ip, fl or le value must be empty'
    description_ru = 'для платежей ИП, ФЛ, И ЮЛ значение должно быть пустым'


class PayerKPPValidationINN10EmptyNotAllowed(PayerKPPValidationError, KPPValidationEmptyNotAllowed, NeedRequiredField):
    description = 'for budget payment with inn length 10 kpp empty value is not allowed'
    description_ru = 'для платежей в бюджет с длиной ИНН равной 10 значение КПП должно быть заполнено'


class PayerKPPValidationINN12OnlyEmptyError(PayerKPPValidationOnlyEmptyError, IncorrectData):
    description = 'for budget with inn length 12 kpp only empty allowed'
    description_ru = 'для платежей в бюджет с длиной ИНН равной 12 значение КПП должно быть пустым'


class PayerKPPValidationINN5EmptyNotAllowed(PayerKPPValidationError, KPPValidationEmptyNotAllowed, NeedRequiredField):
    description = 'for budget payment with inn length 5 kpp empty value is not allowed'
    description_ru = 'для платежей в бюджет с длиной ИНН равной 5 значение КПП должно быть заполнено'


class ReceiverKPPValidationError(KPPValidationError):
    target = 'receiver kpp'
    target_ru = 'КПП получателя'
    description = 'base error'
    description_ru = 'базовая ошибка'


class ReceiverKPPValidationOnlyEmptyError(ReceiverKPPValidationError, KPPValidationOnlyEmptyError, IncorrectData):
    description = 'for ip or fl only empty allowed'
    description_ru = 'для платежей ИП и ФЛ значение должно быть пустым'


class ReceiverKPPValidationEmptyNotAllowed(ReceiverKPPValidationError, KPPValidationEmptyNotAllowed):
    description = 'for fns, customs, budget other empty value is not allowed'
    description_ru = 'для платежей в бюджет пустое значение недопустимо'


class ReceiverKPPValidationStartsWithZeros(ReceiverKPPValidationError, IncorrectData):
    description = 'for fns, customs, budget other kpp cannot starts with "00"'
    description_ru = 'для платежей в бюджет значение не может начинаться с "00"'


class ReceiverKPPValidationFNS(ReceiverKPPValidationError, IncorrectData):
    description = f'for fns kpp can only be {FNS_KPP}'
    description_ru = f'для платежей в фнс значение может быть только {FNS_KPP}'


class ReceiverKPPValidationFTS(ReceiverKPPValidationError, IncorrectData):
    description = f'for customs kpp can only be {FTS_KPP}'
    description_ru = f'для платежей в таможню значение может быть только {FTS_KPP}'


class CBCValidationError(VityaDescribedError, PydanticValueError):
    target = 'CBC'
    target_ru = 'КБК'
    description = 'base error'
    description_ru = 'базовая ошибка'


class CBCValidationTypeError(CBCValidationError, PydanticTypeError):
    description = 'must be str'
    description_ru = 'должен быть строкой'


class CBCValidationEmptyNotAllowed(CBCValidationError, NeedRequiredField):
    description = 'for fns or customs empty value is not allowed'
    description_ru = 'для платежей в ФНС или таможню значение не должно быть пустым'


class CBCValidationValueLenError(CBCValidationError, ExactFieldLenError):
    description = 'must be 20 digits'
    description_ru = 'должен состоять из 20 цифр'
    required_len = 20


class CBCValidationValueDigitsOnlyError(CBCValidationError, IncorrectData):
    description = 'only digits are allowed'
    description_ru = 'должен состоять только из цифр'


class CBCValidationValueCannotZerosOnly(CBCValidationError, IncorrectData):
    description = 'cannot contain only zeros'
    description_ru = 'не может состоять только из нулей'


class OKTMOValidationEmptyNotAllowed(OKTMOValidationError, NeedRequiredField):
    target = 'oktmo'
    target_ru = 'ОКТМО'
    description = 'empty value is not allowed'
    description_ru = 'значение не должно быть пустым'


class OKTMOValidationFNSEmptyNotAllowed(OKTMOValidationEmptyNotAllowed, NeedRequiredField):
    description = 'invalid oktmo: for fns with payer status = "02" empty value is not allowed'
    description_ru = 'для платежей в фнс со статусом плательщика "02" значение не может быть пустым'


class OKTMOValidationZerosNotAllowed(OKTMOValidationError, IncorrectData):
    description = 'cannot be all zeros'
    description_ru = 'не может состоять полностью из нулей'


class OKTMOValidationFTS(OKTMOValidationError, IncorrectData):
    description = f'for customs payments oktmo can be only {FTS_OKTMO}'
    description_ru = f'для платежей в таможню октмо может быть только {FTS_OKTMO}'


class ReasonValidationError(VityaDescribedError, PydanticValueError):
    target = 'reason'
    target_ru = 'Основание платежа'
    description = 'base error'
    description_ru = 'базовая ошибка'


class ReasonValidationTypeError(ReasonValidationError, PydanticTypeError):
    description = 'must be str'
    description_ru = 'должно быть строкой'


class ReasonValidationValueLenError(ReasonValidationError, IncorrectLen):
    description = 'must be 2 chars'
    description_ru = 'должно состоять из 2 символов'


class ReasonValidationValueErrorCustoms(ReasonValidationError, IncorrectData):
    description = f'for customs payment value must be in {CUSTOMS_REASONS}'
    description_ru = f'для платежей в таможню значение должно быть одним из {CUSTOMS_REASONS}'


class ReasonValidationValueErrorFNS(ReasonValidationError, IncorrectData):
    description = 'for fns payment value must be either 0 or empty'
    description_ru = 'для платежей в ФНС значение должно быть 0 или пустым'


class TaxPeriodValidationError(VityaDescribedError, PydanticValueError):
    target = 'tax period'
    target_ru = 'Периодичность платежа / Код таможенного органа'
    description = 'base error'
    description_ru = 'базовая ошибка'


class TaxPeriodValidationTypeError(TaxPeriodValidationError, PydanticTypeError):
    description = 'must be str'
    description_ru = 'должно быть строкой'


class TaxPeriodValidationEmptyNotAllowed(TaxPeriodValidationError):
    description = 'empty is not allowed'
    description_ru = 'значение не должно быть пустым'


class TaxPeriodValidationValueLenError(TaxPeriodValidationError):
    description = 'invalid len'
    description_ru = 'некорректная длина'


class TaxPeriodValidationBOValueLenError(TaxPeriodValidationValueLenError, IncorrectLen):
    description = 'for bo must be 10'
    description_ru = 'для иных платежей в бюджет длина должны быть равна 10'


class TaxPeriodValidationCustomsEmptyNotAllowed(TaxPeriodValidationEmptyNotAllowed, NeedRequiredField):
    pass


class TaxPeriodValidationCustomsValueLenError(TaxPeriodValidationValueLenError, ExactFieldLenError):
    description = 'for customs len must be 8'
    description_ru = 'для платежей в таможню длина должна быть 8 символов'
    required_len = 8


class TaxPeriodValidationFNS02EmptyNotAllowed(TaxPeriodValidationError, NeedRequiredField):
    description = 'for fns with payer status = "02" empty is not allowed'
    description_ru = 'для платежей в ФНС со статусом плательщика "02" пустое значение недопустимо'


class TaxPeriodValidationFNS01OnlyEmpty(TaxPeriodValidationError, IncorrectData):
    description = 'for fns with payer status = "01" or "13" only empty allowed'
    description_ru = 'для платежей в ФНС со статусом плательщика "01" или "13" значение должно быть пустым'


class TaxPeriodValidationFNSValueLenError(TaxPeriodValidationValueLenError, ExactFieldLenError):
    description = 'for fns len must be 10 chars'
    description_ru = 'для платежей в ФНС длина значения должна быть 10 символов'
    required_len = 10


class DocumentNumberValidationError(VityaDescribedError, PydanticValueError):
    target = 'document number'
    target_ru = 'Номер документа'
    description = 'base error'
    description_ru = 'базовая ошибка'


class DocumentNumberValidationTypeError(DocumentNumberValidationError, PydanticTypeError):
    description = 'must be a str'
    description_ru = 'должен быть строкой'


class DocumentNumberValidationOnlyEmptyError(DocumentNumberValidationError):
    description = 'only empty allowed'
    description_ru = 'должно быть пустым'


class DocumentNumberValidationEmptyNotAllowed(DocumentNumberValidationError):
    description = 'empty is not allowed'
    description_ru = 'значение не должно быть пустым'


class DocumentNumberValidationFNSOnlyEmptyError(DocumentNumberValidationOnlyEmptyError, IncorrectData):
    description = 'for fns only empty allowed'
    description_ru = 'для платежей в ФНС значение должно быть пустым'


class DocumentNumberValidationBOEmptyNotAllowed(DocumentNumberValidationEmptyNotAllowed):
    description = 'for bo with payer status = "24", empty payer inn and empty uin empty is not allowed'
    description_ru = (
        'для иных платежей в бюджет со статусом плательщика "24", пустое значение ИНН плательщика и УИН недопустимо'
    )


class DocumentNumberValidationBOPayerStatus33OnlyEmptyError(DocumentNumberValidationOnlyEmptyError, IncorrectData):
    description = (
        'for bo with payer status = "33" , must be empty'
    )
    description_ru = (
        'для иных платежей в бюджет со статусом плательщика = «33» поле документ должно быть обязательно пустым'
    )


class DocumentNumberValidationBOOnlyEmptyError(DocumentNumberValidationOnlyEmptyError, IncorrectData):
    description = (
        'for bo with receiver account starts with "03212",'
        ' payer status = "31", uin is not empty - empty value is not allowed'
    )
    description_ru = (
        'для иных платежей в бюджет счет получателя должен начинаться с "03212",'
        ' статус плательщика должен быть "31, а УИН быть заполнен'
    )


class DocumentNumberValidationBOValueError(DocumentNumberValidationError, IncorrectData):
    description = (
        f'for bo with tax payer status 24 first two chars should be '
        f'in {DOCUMENT_NUMBERS}, and third is equal to ";"'
    )
    description_ru = (
        f'для иных платежей в бюджет со статусом плательщика 24 первые два символы должны быть в {DOCUMENT_NUMBERS},'
        f' а третий символ должен быть равен ";"'
    )


class DocumentNumberValidationBOValueLenError(DocumentNumberValidationError, IncorrectLen):
    description = 'for bo value len max 15'
    description_ru = 'для иных платежей в бюджет максимальная длина должна быть равна 15'


class DocumentNumberValidationCustoms00ValueError(DocumentNumberValidationError, IncorrectData):
    description = 'for customs with reason = "00" value have to be "00" or "0"'
    description_ru = 'для платежей в таможню основание платежа должно быть равно "00", а номер должен быть "00" или "0"'


class DocumentNumberValidationCustomsValueLen7Error(DocumentNumberValidationError, IncorrectLen):
    description = 'for customs with reason in {"ПК", "УВ", "ТГ", "ТБ", "ТД", "ПВ"} value len max 7'
    description_ru = (
        'для платежей в таможню с основанием платежа в {"ПК", "УВ", "ТГ", "ТБ", "ТД", "ПВ"}'
        ' длина значения должна быть не более 7 символов'
    )


class DocumentNumberValidationCustomsValueLen15Error(DocumentNumberValidationError, IncorrectLen):
    description = 'for customs with reason in {"ИЛ", "ИН", "ПБ", "КЭ"} value len from 1 to 15 chars'
    description_ru = (
        'для платежей в таможню с основанием платежа в {"ИЛ", "ИН", "ПБ", "КЭ"}'
        ' длина значения должна быть от 1 до 15 символов'
    )


class DocumentDateValidationError(VityaDescribedError, PydanticValueError):
    target = 'document date'
    target_ru = 'Дата документа'
    description = 'base error'
    description_ru = 'базовая ошибка'


class DocumentDateValidationTypeError(DocumentDateValidationError, PydanticTypeError):
    description = 'must be str'
    description_ru = 'должна быть строкой'


class DocumentDateValidationFNSOnlyEmptyError(DocumentDateValidationError, IncorrectData):
    description = 'for fns only empty allowed'
    description_ru = 'для платежей в ФНС значение должно быть пустым'


class DocumentDateValidationCustomsLenError(DocumentDateValidationError, ExactFieldLenError):
    description = 'for customs value must be equal 10 chars'
    description_ru = 'для платежей в таможню длина значения должна быть 10 символов'
    required_len = 10


class DocumentDateValidationBOLenError(DocumentDateValidationError, IncorrectLen):
    description = 'for bo value max 10 chars'
    description_ru = 'для иных платежей в бюджет значение не должно быть длиннее 10 символов'


class DocumentDateValidationCustomsReasonValueError(DocumentDateValidationError, IncorrectData):
    description = 'for customs with reason = "00" value have to be "0" or "00" or empty'
    description_ru = 'для платежей в таможню c основанием платежа "00" значение должно быть пустым или "0" или "00"'


class BudgetPaymentForThirdPersonError(VityaDescribedError, PydanticValueError):
    target = None
    target_ru = None
    description = 'budget payment can not for third person'
    description_ru = 'платеж в бюджет не может быть за третье лицо'


class TypeOfIncomeValidationError(VityaDescribedError, PydanticValueError):
    target = 'type of income'
    target_ru = 'Код вида дохода'
    description = 'type of income can be only in 1, 2, 3, 4, 5'
    description_ru = 'код вида дохода может быть только 1, 2, 3, 4, 5'


class TypeOfIncomeValidationTypeError(TypeOfIncomeValidationError):
    description = 'must be str'
    description_ru = 'должен быть строкой'


class ReceiverAccountValidationBudgetPayerStatusError(ReceiverAccountValidationError, IncorrectData):
    description = (
        'for budget payment with payer status in 01, 02, 04, 06, 07, 13, 16, 17, 28, 30 '
        'account must start with "03100"'
    )
    description_ru = (
        'для платежей в бюджет со статусом плательщика 01, 02, 04, 06, 07, 13, 16, 17, '
        '28, 30 счет должен начинаться с "03100"'
    )


class ReceiverAccountValidationBudgetOtherPayerStatusError(ReceiverAccountValidationError, IncorrectData):
    description = (
        'for budget other payment with payer status = 31 account must start with "03212"'
    )
    description_ru = (
        'для иных платежей в бюджет со статусом плательщика 31 счет должен начинаться с "03212"'
    )


class TaxPeriodValidationBOValueOnlyOneZeroAllowed(TaxPeriodValidationError, IncorrectData):
    description = 'for bo must contain only one "0"'
    description_ru = 'для иных платежей в бюджет значение не должно состоять из более чем одного "0"'



import re
from datetime import date
from typing import Optional

from vitya.payment_order.enums import PaymentType
from vitya.payment_order.errors import (  # DocumentNumberValidationBOValueError,
    AccountValidationBICValueError,
    BudgetPaymentForThirdPersonError,
    CBCValidationEmptyNotAllowed,
    DocumentDateValidationBOLenError,
    DocumentDateValidationCustomsLenError,
    DocumentDateValidationCustomsReasonValueError,
    DocumentDateValidationFNSOnlyEmptyError,
    DocumentNumberValidationBOEmptyNotAllowed,
    DocumentNumberValidationBOOnlyEmptyError,
    DocumentNumberValidationBOPayerStatus33OnlyEmptyError,
    DocumentNumberValidationBOValueError,
    DocumentNumberValidationBOValueLenError,
    DocumentNumberValidationCustoms00ValueError,
    DocumentNumberValidationCustomsValueLen7Error,
    DocumentNumberValidationCustomsValueLen15Error,
    DocumentNumberValidationFNSOnlyEmptyError,
    OKTMOValidationEmptyNotAllowed,
    OKTMOValidationFNSEmptyNotAllowed,
    OKTMOValidationFTS,
    OKTMOValidationZerosNotAllowed,
    OperationKindValidationBudgetValueError,
    PayerINNValidationCustomsLen10Error,
    PayerINNValidationCustomsLen12Error,
    PayerINNValidationEmptyNotAllowedError,
    PayerINNValidationStartWithZerosError,
    PayerKPPValidationINN5EmptyNotAllowed,
    PayerKPPValidationINN10EmptyNotAllowed,
    PayerKPPValidationINN12OnlyEmptyError,
    PayerStatusValidationCustoms05NotAllowedError,
    PayerStatusValidationCustomsIncorrectDataError,
    PayerStatusValidationFNSIncorrectDataError,
    PayerStatusValidationNullNotAllowedError,
    PayerStatusValidationOtherIncorrectDataError,
    PaymentTypeValueError,
    PurposeCodeValidationChameleonError,
    PurposeCodeValidationFlError,
    PurposeCodeValidationNullError,
    PurposeValidationForThirdPersonError,
    PurposeValidationValueEmptyErrorForNonFNS,
    ReasonValidationValueErrorCustoms,
    ReasonValidationValueErrorFNS,
    ReceiverAccountValidationBICValueError,
    ReceiverAccountValidationBudgetOtherPayerStatusError,
    ReceiverAccountValidationBudgetPayerStatusError,
    ReceiverAccountValidationCustomsValueError,
    ReceiverAccountValidationFNSValueError,
    ReceiverINNValidationChameleonLenError,
    ReceiverINNValidationFLLenError,
    ReceiverINNValidationIPLenError,
    ReceiverINNValidationLELenError,
    ReceiverINNValidationNonEmptyError,
    ReceiverKPPValidationEmptyNotAllowed,
    ReceiverKPPValidationFNS,
    ReceiverKPPValidationFTS,
    ReceiverKPPValidationOnlyEmptyError,
    ReceiverKPPValidationStartsWithZeros,
    TaxPeriodValidationBOValueLenError,
    TaxPeriodValidationBOValueOnlyOneZeroAllowed,
    TaxPeriodValidationCustomsEmptyNotAllowed,
    TaxPeriodValidationCustomsValueLenError,
    TaxPeriodValidationFNS01OnlyEmpty,
    TaxPeriodValidationFNS02EmptyNotAllowed,
    TaxPeriodValidationFNSValueLenError,
    UINValidationBONotEmpty,
    UINValidationFNSNotValueZeroError,
    UINValidationFNSOrFTSLenError,
    UINValidationFNSValueZeroError,
    UINValidationValueBudget33PayerStatusIncorrectLength,
    UINValidationValueZeroError,
)
from vitya.payment_order.fields import (
    CBC,
    UIN,
    AccountNumber,
    DocumentDate,
    DocumentNumber,
    ForThirdPerson,
    OperationKind,
    PayerINN,
    PayerKPP,
    PayerStatus,
    Purpose,
    PurposeCode,
    Reason,
    ReceiverAccountNumber,
    ReceiverBIC,
    ReceiverINN,
    ReceiverKPP,
    TaxPeriod,
)
from vitya.payment_order.payments.constants import (
    CHANGE_YEAR,
    CUSTOMS_REASONS,
    CUSTOMS_RECEIVER_ACCOUNT_NUMBER,
    DOCUMENT_NUMBERS,
    FNS_KPP,
    FNS_RECEIVER_ACCOUNT_NUMBER,
    FNS_TAX_PAYER_STATUSES,
    FTS_KPP,
    FTS_OKTMO,
    FTS_TAX_PAYER_STATUSES,
    OTHER_OKTMO_RECEIVER_ACCOUNT_PREFIXES,
    OTHER_OKTMO_RECEIVER_ACCOUNT_PREFIXES_2,
)
from vitya.pydantic_fields import BIC, OKTMO


def check_account_by_bic(
    account_number: AccountNumber,
    bic: BIC,
) -> None:
    _sum = 0
    for c, v in zip(bic[-3:] + account_number, [7, 1, 3] * 8):
        _sum += int(c) * v
    if _sum % 10 != 0:
        raise AccountValidationBICValueError


def check_receiver_account(
    value: ReceiverAccountNumber,
    payment_type: PaymentType,
    receiver_bic: ReceiverBIC,
) -> str:
    if payment_type == PaymentType.FNS:
        if value != FNS_RECEIVER_ACCOUNT_NUMBER:
            raise ReceiverAccountValidationFNSValueError
    elif not payment_type.is_budget:
        try:
            check_account_by_bic(account_number=value, bic=receiver_bic)
        except AccountValidationBICValueError as e:
            raise ReceiverAccountValidationBICValueError from e
    return value


def check_receiver_account_with_payment_type(
    value: ReceiverAccountNumber,
    payment_type: PaymentType,
) -> str:
    if payment_type == PaymentType.CUSTOMS and value != CUSTOMS_RECEIVER_ACCOUNT_NUMBER:
        raise ReceiverAccountValidationCustomsValueError
    return value


def check_receiver_account_with_payment_type_and_payer_status(
    value: ReceiverAccountNumber,
    payment_type: PaymentType,
    payer_status: Optional[PayerStatus],
) -> str:
    if (
        payment_type.is_budget
        and payer_status in ['01', '02', '04', '06', '07', '13', '16', '17', '28', '30']
        and not value.startswith('03100')
    ):
        raise ReceiverAccountValidationBudgetPayerStatusError

    elif (
        payment_type == PaymentType.BUDGET_OTHER
        and payer_status == '31'
        and not value.startswith('03212')
    ):
        raise ReceiverAccountValidationBudgetOtherPayerStatusError

    return value


def check_operation_kind(
    value: OperationKind,
    payment_type: PaymentType
) -> OperationKind:
    if payment_type.is_budget and value not in {'01', '02', '06'}:
        raise OperationKindValidationBudgetValueError
    return value


def check_purpose_code(
    value: Optional[PurposeCode],
    payment_type: PaymentType,
) -> Optional[PurposeCode]:
    if payment_type not in (PaymentType.FL, PaymentType.CHAMELEON):
        if value is not None:
            raise PurposeCodeValidationNullError
        return None
    if value is not None and value not in {1, 2, 3, 4, 5}:
        if payment_type == PaymentType.FL:
            raise PurposeCodeValidationFlError
        raise PurposeCodeValidationChameleonError
    return value


def check_uin(
    value: Optional[UIN],
    receiver_account: ReceiverAccountNumber,
    payment_type: PaymentType,
    payer_status: Optional[PayerStatus],
    payer_inn: Optional[PayerINN],
) -> Optional[UIN]:
    if not payment_type.is_budget:
        return None

    if payer_status == '31' and value is None:
        raise UINValidationValueZeroError

    if (
        payment_type.is_budget and payer_status == '33'
        and (not value or len(value) not in [20, 25] or value == '0' * len(value))
    ):
        raise UINValidationValueBudget33PayerStatusIncorrectLength

    if value is not None and payment_type in {PaymentType.FNS, PaymentType.CUSTOMS} and len(value) not in [25, 20]:
        raise UINValidationFNSOrFTSLenError

    if payment_type == PaymentType.BUDGET_OTHER:
        if receiver_account.startswith('03212') and value is None:
            raise UINValidationBONotEmpty
        return value

    if payment_type == PaymentType.FNS:
        if payer_status == '13' and payer_inn is None and value is None:
            raise UINValidationFNSValueZeroError
        if payer_status == '02' and date.today().year < CHANGE_YEAR:
            if value is not None:
                raise UINValidationFNSNotValueZeroError
            return value
    return value


def check_purpose(
    value: Optional[Purpose],
    payment_type: PaymentType,
) -> Optional[Purpose]:
    if payment_type == PaymentType.FNS and not value:
        return Purpose('0')

    if payment_type != PaymentType.FNS and not value:
        raise PurposeValidationValueEmptyErrorForNonFNS

    return value


def check_payer_inn(
    value: Optional[PayerINN],
    payment_type: PaymentType,
    payer_status: Optional[PayerStatus],
    for_third_person: ForThirdPerson,
) -> Optional[PayerINN]:
    if not payment_type.is_budget:
        return value

    if value is None:
        if payment_type == PaymentType.BUDGET_OTHER:
            return None
        elif payment_type == PaymentType.FNS:
            return None
        elif payment_type == PaymentType.CUSTOMS and payer_status == '30':
            return None
        raise PayerINNValidationEmptyNotAllowedError

    if payment_type == PaymentType.CUSTOMS:
        if payer_status == '06' and for_third_person and len(value) != 10:
            raise PayerINNValidationCustomsLen10Error

        if payer_status in {'16', '17'} and len(value) != 12:
            raise PayerINNValidationCustomsLen12Error

    if value.startswith('00'):
        raise PayerINNValidationStartWithZerosError

    return value


def check_payer_inn_with_uin_and_receiver_account(
    value: Optional[PayerINN],
    payment_type: PaymentType,
    payer_status: Optional[PayerStatus],
    receiver_account: Optional[ReceiverAccountNumber],
    uin: Optional[UIN],
) -> Optional[PayerINN]:
    if value is None and payment_type == PaymentType.FNS:
        if not (
            payer_status == '13'
            and receiver_account is not None
            and receiver_account.startswith('03100')
            and uin is not None
            and len(uin) in [20, 25]
        ):
            raise PayerINNValidationEmptyNotAllowedError

    return None


def check_receiver_inn(
    value: Optional[ReceiverINN],
    payment_type: PaymentType,
) -> Optional[ReceiverINN]:
    if payment_type in [
        PaymentType.CUSTOMS,
        PaymentType.FNS,
        PaymentType.BUDGET_OTHER,
        PaymentType.LE,
        PaymentType.IP
    ] and value is None:
        raise ReceiverINNValidationNonEmptyError
    if payment_type in [
        PaymentType.CUSTOMS,
        PaymentType.FNS,
        PaymentType.BUDGET_OTHER,
        PaymentType.LE
    ] and (value is None or len(value) != 10):
        raise ReceiverINNValidationLELenError

    if payment_type == PaymentType.IP:
        if value is None or len(value) != 12:
            raise ReceiverINNValidationIPLenError
    elif payment_type == PaymentType.FL:
        if value is not None and len(value) != 12:
            raise ReceiverINNValidationFLLenError
    elif payment_type == PaymentType.CHAMELEON:
        if value is not None and len(value) not in (10, 12):
            raise ReceiverINNValidationChameleonLenError

    return value


def check_payment_type_and_for_third_person(
    payment_type: PaymentType,
    for_third_person: ForThirdPerson,
) -> None:
    if for_third_person and not payment_type.is_budget:
        raise BudgetPaymentForThirdPersonError


def check_purpose_for_third_person(
    value: Optional[Purpose],
    for_third_person: ForThirdPerson,
) -> Optional[Purpose]:
    if not for_third_person:
        return value

    if not value or not re.match(r'^\d+\/\/[a-zA-Zа-яА-ЯёЁ\s\W]+\/\/[а-яА-ЯёЁ\s!-~]*$', value):
        raise PurposeValidationForThirdPersonError

    return value


def check_payer_status(
    value: Optional[PayerStatus],
    payment_type: PaymentType,
    for_third_person: Optional[ForThirdPerson],
) -> Optional[PayerStatus]:
    if not payment_type.is_budget:
        return None

    if value is None:
        raise PayerStatusValidationNullNotAllowedError

    if payment_type == PaymentType.FNS and value not in FNS_TAX_PAYER_STATUSES:
        raise PayerStatusValidationFNSIncorrectDataError

    if payment_type == PaymentType.CUSTOMS and value not in FTS_TAX_PAYER_STATUSES:
        raise PayerStatusValidationCustomsIncorrectDataError

    if payment_type == PaymentType.BUDGET_OTHER and value in FNS_TAX_PAYER_STATUSES + FTS_TAX_PAYER_STATUSES:
        raise PayerStatusValidationOtherIncorrectDataError

    if payment_type == PaymentType.CUSTOMS and for_third_person == False and value == '06':  # noqa
        raise PayerStatusValidationCustoms05NotAllowedError

    return value


def check_payer_kpp(
    value: Optional[PayerKPP],
    payment_type: PaymentType,
    payer_inn: Optional[PayerINN],
    payer_status: Optional[PayerStatus],
) -> Optional[PayerKPP]:
    if not payment_type.is_budget:
        return None

    if payer_inn is not None and len(payer_inn) == 5 and value is None:
        raise PayerKPPValidationINN5EmptyNotAllowed
    elif payer_inn is not None and len(payer_inn) == 10 and value is None and payer_status != '01':
        raise PayerKPPValidationINN10EmptyNotAllowed
    elif payer_inn is not None and len(payer_inn) == 12 and value is not None:
        raise PayerKPPValidationINN12OnlyEmptyError
    return value


def check_receiver_kpp(
    value: Optional[ReceiverKPP],
    payment_type: PaymentType,
) -> Optional[ReceiverKPP]:
    if payment_type in {PaymentType.FL, PaymentType.IP}:
        if value is not None:
            raise ReceiverKPPValidationOnlyEmptyError
        return None
    if payment_type in {PaymentType.LE, PaymentType.CHAMELEON}:
        return value
    if value is None:
        raise ReceiverKPPValidationEmptyNotAllowed
    if value.startswith('00'):
        raise ReceiverKPPValidationStartsWithZeros
    if payment_type == PaymentType.FNS and value != FNS_KPP:
        raise ReceiverKPPValidationFNS
    if payment_type == PaymentType.CUSTOMS and value != FTS_KPP:
        raise ReceiverKPPValidationFTS
    return value


def check_cbc(
    value: Optional[CBC],
    payment_type: PaymentType,
) -> Optional[CBC]:
    if not payment_type.is_budget:
        return None

    if payment_type == PaymentType.BUDGET_OTHER and value is None:
        return None

    if payment_type in {PaymentType.FNS, PaymentType.CUSTOMS} and value is None:
        raise CBCValidationEmptyNotAllowed

    return value


def check_oktmo(
    value: Optional[OKTMO],
    payment_type: PaymentType,
) -> Optional[OKTMO]:
    if not payment_type.is_budget:
        return None

    if payment_type == PaymentType.CUSTOMS and value != FTS_OKTMO:
        raise OKTMOValidationFTS

    if payment_type in {PaymentType.FNS, PaymentType.BUDGET_OTHER} and value is None:
        return None

    if value is None:
        raise OKTMOValidationEmptyNotAllowed

    if all(c == '0' for c in value):
        raise OKTMOValidationZerosNotAllowed
    return value


def check_oktmo_with_payer_status(
    value: Optional[OKTMO],
    payment_type: PaymentType,
    payer_status: Optional[PayerStatus],
) -> Optional[OKTMO]:
    if payment_type == PaymentType.FNS and payer_status in {'01', '13'} and value is None:
        return None

    if (
        payment_type == PaymentType.FNS and
        payer_status == '02' and
        value is None and
        date.today().year < CHANGE_YEAR
    ):
        raise OKTMOValidationFNSEmptyNotAllowed

    return value


def check_oktmo_with_receiver_account_number(
    value: Optional[OKTMO],
    payment_type: PaymentType,
    receiver_account_number: ReceiverAccountNumber,
) -> Optional[OKTMO]:
    if payment_type == PaymentType.BUDGET_OTHER and (value is None or value == '0'):
        if (
            receiver_account_number[:5] in OTHER_OKTMO_RECEIVER_ACCOUNT_PREFIXES
            or (
                receiver_account_number[:5] in OTHER_OKTMO_RECEIVER_ACCOUNT_PREFIXES_2
                and receiver_account_number[13] == '4'
            )
        ):
            return None
        raise OKTMOValidationEmptyNotAllowed

    return value


def check_reason(
    value: Optional[Reason],
    payment_type: PaymentType,
) -> Optional[Reason]:
    if not payment_type.is_budget:
        return None

    if payment_type.is_budget and (value is None or value == '0'):
        return None
    if payment_type == PaymentType.CUSTOMS and value not in CUSTOMS_REASONS:
        raise ReasonValidationValueErrorCustoms
    if payment_type == PaymentType.FNS and value and value != '0':
        raise ReasonValidationValueErrorFNS
    return value


def check_tax_period(
    value: Optional[TaxPeriod],
    payment_type: PaymentType,
    payer_status: Optional[PayerStatus],
) -> Optional[TaxPeriod]:
    if not payment_type.is_budget:
        return None

    if payment_type == PaymentType.BUDGET_OTHER:
        if value is None:
            return None
        elif len(value) > 10:
            raise TaxPeriodValidationBOValueLenError
        elif value == '0' * len(value):
            raise TaxPeriodValidationBOValueOnlyOneZeroAllowed
        return value
    elif payment_type == PaymentType.CUSTOMS:
        if value is None:
            raise TaxPeriodValidationCustomsEmptyNotAllowed
        elif len(value) != 8:
            raise TaxPeriodValidationCustomsValueLenError
        return value
    else:
        if (
            payer_status == '02' and
            value is None and
            date.today().year < CHANGE_YEAR
        ):
            raise TaxPeriodValidationFNS02EmptyNotAllowed
        if payer_status in {'01', '13'}:
            if value is not None:
                raise TaxPeriodValidationFNS01OnlyEmpty
            return None

        if payer_status == '02' and len(value or '') != 10 and date.today().year < CHANGE_YEAR:
            raise TaxPeriodValidationFNSValueLenError
        return value


def check_document_number(
    value: Optional[DocumentNumber],
    payment_type: PaymentType,
    reason: Optional[Reason],
    payer_status: Optional[PayerStatus],
    uin: Optional[UIN],
    payer_inn: Optional[PayerINN],
) -> Optional[DocumentNumber]:
    if not payment_type.is_budget:
        return None

    if payer_status == '31':
        if value is not None:
            raise DocumentNumberValidationBOOnlyEmptyError
        return None

    if payment_type == PaymentType.FNS:
        if value is not None:
            raise DocumentNumberValidationFNSOnlyEmptyError
        return None
    elif payment_type == PaymentType.BUDGET_OTHER:
        if payer_status == '33' and date.today().year >= CHANGE_YEAR:
            if value is not None:
                raise DocumentNumberValidationBOPayerStatus33OnlyEmptyError
            return None
        if payer_status == '24' and payer_inn is None and uin is None and value is None:
            raise DocumentNumberValidationBOEmptyNotAllowed
        if value is not None:
            if len(value) > 15:
                raise DocumentNumberValidationBOValueLenError
            if (
                payer_status == '24' and (
                    len(value) < 3 or value[2] != ';' or value[:2] not in DOCUMENT_NUMBERS
                )
            ):
                raise DocumentNumberValidationBOValueError
        return value
    elif payment_type == PaymentType.CUSTOMS:
        if reason == '00' and value is not None and value not in ['00', '0']:
            raise DocumentNumberValidationCustoms00ValueError
        if reason in {'ПК', 'УВ', 'ТГ', 'ТБ', 'ТД', 'ПВ'} and value and len(value) > 7:
            raise DocumentNumberValidationCustomsValueLen7Error
        if (
            reason in {'ИЛ', 'ИН', 'ПБ', 'КЭ'}
            and (
                value is None or len(value) > 15
            )
        ):
            raise DocumentNumberValidationCustomsValueLen15Error
        return value
    raise PaymentTypeValueError(payment_type=payment_type)  # pragma: no cover


def check_document_date(
    value: Optional[DocumentDate],
    payment_type: PaymentType,
) -> Optional[DocumentDate]:
    if not payment_type.is_budget:
        return None

    if payment_type == PaymentType.FNS:
        if value is not None and value != '0':
            raise DocumentDateValidationFNSOnlyEmptyError
        return None
    else:
        if value is None:
            return None
        if payment_type == PaymentType.CUSTOMS:
            if len(value) != 10:
                raise DocumentDateValidationCustomsLenError
            return value
        else:
            if len(value) > 10:
                raise DocumentDateValidationBOLenError
            return value


def check_document_date_with_reason(
    value: Optional[DocumentDate],
    payment_type: PaymentType,
    reason: Reason,
) -> Optional[DocumentDate]:
    if payment_type == PaymentType.CUSTOMS and reason == '00':
        if not (value is None or value in ['0', '00']):
            raise DocumentDateValidationCustomsReasonValueError
    return value
